/*
ucspi -- UCSPI-related utilities
Copyright (C) 2021-2023 Sergey Matveev <stargrave@stargrave.org>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

package ucspi

import (
	"errors"
	"net"
	"os"
	"time"
)

type Addr struct {
	ip   string
	port string
}

func (addr *Addr) Network() string { return "tcp" }

func (addr *Addr) String() string { return addr.ip + ":" + addr.port }

type Conn struct {
	R *os.File
	W *os.File
}

func NewConn(r, w *os.File) (*Conn, error) {
	if r == nil {
		return nil, errors.New("no R file descriptor")
	}
	if w == nil {
		return nil, errors.New("no W file descriptor")
	}
	return &Conn{R: r, W: w}, nil
}

func (conn *Conn) Read(b []byte) (int, error) {
	return conn.R.Read(b)
}

func (conn *Conn) Write(b []byte) (int, error) { return conn.W.Write(b) }

func (conn *Conn) Close() error {
	errR := conn.R.Close()
	errW := conn.W.Close()
	if errR != nil {
		return errR
	}
	return errW
}

func (conn *Conn) LocalAddr() net.Addr {
	return &Addr{ip: os.Getenv("TCPLOCALIP"), port: os.Getenv("TCPLOCALPORT")}
}

func (conn *Conn) RemoteAddr() net.Addr {
	return &Addr{ip: os.Getenv("TCPREMOTEIP"), port: os.Getenv("TCPREMOTEPORT")}
}

func (conn *Conn) SetDeadline(t time.Time) error {
	if err := conn.R.SetReadDeadline(t); err != nil {
		return err
	}
	return conn.W.SetWriteDeadline(t)
}

func (conn *Conn) SetReadDeadline(t time.Time) error {
	return conn.R.SetReadDeadline(t)
}

func (conn *Conn) SetWriteDeadline(t time.Time) error {
	return conn.W.SetWriteDeadline(t)
}
